/*
 * @Author: renjie.yin
 * @Date: 2022-06-07 14:18:09
 * @LastEditors: renjie.yin
 * @LastEditTime: 2022-06-07 19:49:34
 * @Description: 
 */
import 'dart:math' as math;

List<double> newArrayOfZeros(int n) {
  return List<double>.filled(n, 0);
}

void transform(List<double> real, List<double> imag) {
  var n = real.length;
  if (n != imag.length) throw new RangeError("实部虚部长度不匹配");
  if (n == 0)
    return;
  else if ((n & (n - 1)) == 0)
    transformRadix2(real, imag);
  else
    transformBluestein(real, imag);
}

void transformRadix2(List<double> real, List<double> imag) {
  var n = real.length;
  if (n != imag.length) throw new RangeError("实部虚部长度不匹配");
  if (n == 1) return;
  var levels = -1;
  for (var i = 0; i < 32; i++) {
    if (1 << i == n) {
      levels = i;
      break;
    }
  }
  if (levels == -1) throw new RangeError("长度非2的n次方");

  var cosTable = List<double>.filled((n / 2).floor(), 0);
  var sinTable = List<double>.filled((n / 2).floor(), 0);

  for (var i = 0; i < (n / 2).floor(); i++) {
    cosTable[i] = math.cos((2 * math.pi * i) / n);
    sinTable[i] = math.sin((2 * math.pi * i) / n);
  }

  for (var i = 0; i < n; i++) {
    var j = reverseBits(i, levels);
    if (j > i) {
      var temp = real[i];
      real[i] = real[j];
      real[j] = temp;
      temp = imag[i];
      imag[i] = imag[j];
      imag[j] = temp;
    }
  }

  for (var size = 2; size <= n; size *= 2) {
    var halfsize = (size / 2).floor();
    var tablestep = (n / size).floor();
    for (var i = 0; i < n; i += size) {
      for (var j = i, k = 0; j < i + halfsize; j++, k += tablestep) {
        var l = j + halfsize;
        var tpre = real[l] * cosTable[k] + imag[l] * sinTable[k];
        var tpim = -real[l] * sinTable[k] + imag[l] * cosTable[k];
        real[l] = real[j] - tpre;
        imag[l] = imag[j] - tpim;
        real[j] += tpre;
        imag[j] += tpim;
      }
    }
  }
}

int reverseBits(int val, int width) {
  int result = 0;
  for (var i = 0; i < width; i++) {
    result = (result << 1) | (val & 1);
    val >>>= 1;
  }
  return result;
}

void transformBluestein(List<double> real, List<double> imag) {
  var n = real.length;
  if (n != imag.length) throw new RangeError("实部虚部长度不匹配");
  var m = 1;
  while (m < n * 2 + 1) m *= 2;

  var cosTable = List<double>.filled(n, 0);
  var sinTable = List<double>.filled(n, 0);
  for (var i = 0; i < n; i++) {
    var j = (i * i) % (n * 2);
    cosTable[i] = math.cos((math.pi * j) / n);
    sinTable[i] = math.sin((math.pi * j) / n);
  }

  var areal = newArrayOfZeros(m);
  var aimag = newArrayOfZeros(m);
  for (var i = 0; i < n; i++) {
    areal[i] = real[i] * cosTable[i] + imag[i] * sinTable[i];
    aimag[i] = -real[i] * sinTable[i] + imag[i] * cosTable[i];
  }
  var breal = newArrayOfZeros(m);
  var bimag = newArrayOfZeros(m);
  breal[0] = cosTable[0];
  bimag[0] = sinTable[0];
  for (var i = 1; i < n; i++) {
    breal[i] = breal[m - i] = cosTable[i];
    bimag[i] = bimag[m - i] = sinTable[i];
  }

  var creal = List<double>.filled(m, 0);
  var cimag = List<double>.filled(m, 0);
  convolveComplex(areal, aimag, breal, bimag, creal, cimag);

  for (var i = 0; i < n; i++) {
    real[i] = creal[i] * cosTable[i] + cimag[i] * sinTable[i];
    imag[i] = -creal[i] * sinTable[i] + cimag[i] * cosTable[i];
  }
}

void convolveComplex(List<double> xreal, List<double> ximag, List<double> yreal,
    List<double> yimag, List<double> outreal, List<double> outimag) {
  var n = xreal.length;
  if (n != ximag.length ||
      n != yreal.length ||
      n != yimag.length ||
      n != outreal.length ||
      n != outimag.length) throw new RangeError("长度不匹配");

  xreal = xreal.toList();
  ximag = ximag.toList();
  yreal = yreal.toList();
  yimag = yimag.toList();
  transform(xreal, ximag);
  transform(yreal, yimag);

  for (var i = 0; i < n; i++) {
    var temp = xreal[i] * yreal[i] - ximag[i] * yimag[i];
    ximag[i] = ximag[i] * yreal[i] + xreal[i] * yimag[i];
    xreal[i] = temp;
  }
  inverseTransform(xreal, ximag);

  for (var i = 0; i < n; i++) {
    outreal[i] = xreal[i] / n;
    outimag[i] = ximag[i] / n;
  }
}

void inverseTransform(List<double> real, List<double> imag) {
  transform(imag, real);
}

class Complex {
  final double real;
  final double imaginary;

  Complex({this.real = 0, this.imaginary = 0});

  Complex.ri(this.real, this.imaginary);

  double get magnitude {
    return math.sqrt(real * real + imaginary * imaginary).abs();
  }

  Complex div(Complex b1) {
    var a = real;
    var b = imaginary;

    var c = b1.real;
    var d = b1.imaginary;
    var t, x;

    if (0 == d) {
      return Complex(real: a / c, imaginary: b / c);
    }

    if (a.abs() < d.abs()) {
      x = c / d;
      t = c * x + d;

      return Complex(real: (a * x + b) / t, imaginary: (b * x - a) / t);
    } else {
      x = d / c;
      t = d * x + c;

      return Complex(real: (a + b * x) / t, imaginary: (b - a * x) / t);
    }
  }
}

List<Complex> realFFT(List<double> data) {
  var real = data.toList();
  var imag = List<double>.filled(data.length, 0);
  transform(real, imag);
  List<Complex> res = [];
  for (int i = 0; i < real.length; i++) {
    res.add(Complex.ri(real[i], imag[i]));
  }
  return res;
}

List<double> inverseFft(List<Complex> data) {
  List<double> real = List.filled(data.length, 0);
  List<double> imag = List.filled(data.length, 0);
  for (int i = 0; i < data.length; i++) {
    real[i] = data[i].real;
    imag[i] = data[i].imaginary;
  }
  transform(real, imag);
  final len = real.length;
  final scale = len.toDouble();
  final r = List<double>.filled(len, 0);
  r[0] = real[0] / scale;
  if (len <= 1) return r;
  for (int i = 1; i < len; ++i) {
    r[i] = real[len - i] / scale;
  }
  return r;
}
